@@ -22,43 +22,41 @@
  */
 
 #define _GNU_SOURCE
+#include <unistd.h>
+#include <stdio.h>
+#include <string.h>
+#include <stdlib.h>
+#include <signal.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <grp.h>
-#include <pwd.h>
-#include <signal.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <unistd.h>
-#include <linux/unistd.h>
-#include <sys/mount.h>
 #include <sys/param.h>
 #include <sys/prctl.h>
+#include <sys/mount.h>
 #include <sys/socket.h>
 #include <sys/syscall.h>
 #include <sys/wait.h>
+#include <linux/unistd.h>
+#include <pwd.h>
 
 #if !HAVE_DECL_PR_CAPBSET_DROP
 #define PR_CAPBSET_DROP 24
 #endif
 
-#include "af_unix.h"
+#include "namespace.h"
+#include "log.h"
 #include "attach.h"
 #include "caps.h"
-#include "cgroup.h"
-#include "commands.h"
-#include "conf.h"
 #include "config.h"
-#include "confile.h"
-#include "log.h"
-#include "lxclock.h"
-#include "lxcseccomp.h"
-#include "namespace.h"
 #include "utils.h"
-#include "lsm/lsm.h"
-
+#include "commands.h"
+#include "cgroup.h"
+#include "lxclock.h"
+#include "conf.h"
+#include "lxcseccomp.h"
 #include <lxc/lxccontainer.h>
+#include "lsm/lsm.h"
+#include "confile.h"
 
 #if HAVE_SYS_PERSONALITY_H
 #include <sys/personality.h>
@@ -78,102 +76,6 @@
 
 lxc_log_define(lxc_attach, lxc);
 
-/* /proc/pid-to-str/current\0 = (5 + 21 + 7 + 1) */
-#define __LSMATTRLEN (5 + 21 + 7 + 1)
-static int lsm_openat(int procfd, pid_t pid, int on_exec)
-{
-	int ret = -1;
-	int labelfd = -1;
-	const char *name;
-	char path[__LSMATTRLEN];
-
-	name = lsm_name();
-
-	if (strcmp(name, "nop") == 0)
-		return 0;
-
-	if (strcmp(name, "none") == 0)
-		return 0;
-
-	/* We don't support on-exec with AppArmor */
-	if (strcmp(name, "AppArmor") == 0)
-		on_exec = 0;
-
-	if (on_exec)
-		ret = snprintf(path, __LSMATTRLEN, "%d/attr/exec", pid);
-	else
-		ret = snprintf(path, __LSMATTRLEN, "%d/attr/current", pid);
-	if (ret < 0 || ret >= __LSMATTRLEN)
-		return -1;
-
-	labelfd = openat(procfd, path, O_RDWR);
-	if (labelfd < 0) {
-		SYSERROR("Unable to open file descriptor to set LSM label.");
-		return -1;
-	}
-
-	return labelfd;
-}
-
-static int lsm_set_label_at(int lsm_labelfd, int on_exec, char *lsm_label)
-{
-	int fret = -1;
-	const char* name;
-	char *command = NULL;
-
-	name = lsm_name();
-
-	if (strcmp(name, "nop") == 0)
-		return 0;
-
-	if (strcmp(name, "none") == 0)
-		return 0;
-
-	/* We don't support on-exec with AppArmor */
-	if (strcmp(name, "AppArmor") == 0)
-		on_exec = 0;
-
-	if (strcmp(name, "AppArmor") == 0) {
-		int size;
-
-		command = malloc(strlen(lsm_label) + strlen("changeprofile ") + 1);
-		if (!command) {
-			SYSERROR("Failed to write apparmor profile");
-			goto out;
-		}
-
-		size = sprintf(command, "changeprofile %s", lsm_label);
-		if (size < 0) {
-			SYSERROR("Failed to write apparmor profile");
-			goto out;
-		}
-
-		if (write(lsm_labelfd, command, size + 1) < 0) {
-			SYSERROR("Unable to set LSM label: %s.", command);
-			goto out;
-		}
-		INFO("Set LSM label to: %s.", command);
-	} else if (strcmp(name, "SELinux") == 0) {
-		if (write(lsm_labelfd, lsm_label, strlen(lsm_label) + 1) < 0) {
-			SYSERROR("Unable to set LSM label");
-			goto out;
-		}
-		INFO("Set LSM label to: %s.", lsm_label);
-	} else {
-		ERROR("Unable to restore label for unknown LSM: %s", name);
-		goto out;
-	}
-	fret = 0;
-
-out:
-	free(command);
-
-	if (lsm_labelfd != -1)
-		close(lsm_labelfd);
-
-	return fret;
-}
-
 static struct lxc_proc_context_info *lxc_proc_get_context_info(pid_t pid)
 {
 	struct lxc_proc_context_info *info = calloc(1, sizeof(*info));
@@ -682,7 +584,6 @@ static bool fetch_seccomp(const char *name, const char *lxcpath,
 		struct lxc_proc_context_info *i, lxc_attach_options_t *options)
 {
 	struct lxc_container *c;
-	char *path;
 
 	if (!(options->namespaces & CLONE_NEWNS) || !(options->attach_flags & LXC_ATTACH_LSM))
 		return true;
@@ -691,26 +592,8 @@ static bool fetch_seccomp(const char *name, const char *lxcpath,
 	if (!c)
 		return false;
 	i->container = c;
-
-	/* Initialize an empty lxc_conf */
-	if (!c->set_config_item(c, "lxc.seccomp", "")) {
+	if (!c->lxc_conf)
 		return false;
-	}
-
-	/* Fetch the current profile path over the cmd interface */
-	path = c->get_running_config_item(c, "lxc.seccomp");
-	if (!path) {
-		return true;
-	}
-
-	/* Copy the value into the new lxc_conf */
-	if (!c->set_config_item(c, "lxc.seccomp", path)) {
-		free(path);
-		return false;
-	}
-	free(path);
-
-	/* Attempt to parse the resulting config */
 	if (lxc_read_seccomp_config(c->lxc_conf) < 0) {
 		ERROR("Error reading seccomp policy");
 		return false;
@@ -808,15 +691,9 @@ int lxc_attach(const char* name, const char* lxcpath, lxc_attach_exec_t exec_fun
 	 *        X  <------------------------------------  send 1
 	 *   [add to cgroup, ...]
 	 *    send 2 ------------------------------------>    X
-	 *						[set LXC_ATTACH_NO_NEW_PRIVS]
-	 *        X  <------------------------------------  send 3
-	 *   [open LSM label fd]
-	 *    send 4 ------------------------------------>    X
-	 *   						[set LSM label]
 	 *   close socket                                 close socket
 	 *                                                run program
 	 */
-
 	ret = socketpair(PF_LOCAL, SOCK_STREAM | SOCK_CLOEXEC, 0, ipc_sockets);
 	if (ret < 0) {
 		SYSERROR("could not set up required IPC mechanism for attaching");
@@ -847,7 +724,6 @@ int lxc_attach(const char* name, const char* lxcpath, lxc_attach_exec_t exec_fun
 	}
 
 	if (pid) {
-		int procfd = -1;
 		pid_t to_cleanup_pid = pid;
 
 		/* initial thread, we close the socket that is for the
@@ -862,15 +738,6 @@ int lxc_attach(const char* name, const char* lxcpath, lxc_attach_exec_t exec_fun
 				goto cleanup_error;
 		}
 
-		/* Open /proc before setns() to the containers namespace so we
-		 * don't rely on any information from inside the container.
-		 */
-		procfd = open("/proc", O_DIRECTORY | O_RDONLY | O_CLOEXEC);
-		if (procfd < 0) {
-			SYSERROR("Unable to open /proc.");
-			goto cleanup_error;
-		}
-
 		/* Let the child process know to go ahead */
 		status = 0;
 		ret = lxc_write_nointr(ipc_sockets[0], &status, sizeof(status));
@@ -883,8 +750,7 @@ int lxc_attach(const char* name, const char* lxcpath, lxc_attach_exec_t exec_fun
 		ret = lxc_read_nointr_expect(ipc_sockets[0], &attached_pid, sizeof(attached_pid), NULL);
 		if (ret <= 0) {
 			if (ret != 0)
-				ERROR("error using IPC to receive notification "
-				      "from attached process (1)");
+				ERROR("error using IPC to receive pid of attached process");
 			goto cleanup_error;
 		}
 
@@ -923,44 +789,10 @@ int lxc_attach(const char* name, const char* lxcpath, lxc_attach_exec_t exec_fun
 		status = 2;
 		ret = lxc_write_nointr(ipc_sockets[0], &status, sizeof(status));
 		if (ret <= 0) {
-			ERROR("Error using IPC to notify attached process for "
-			      "initialization (2): %s.", strerror(errno));
+			ERROR("error using IPC to notify attached process for initialization (2)");
 			goto cleanup_error;
 		}
 
-		/* Wait for the (grand)child to tell us that it's ready to set
-		 * up its LSM labels.
-		 */
-		expected = 3;
-		ret = lxc_read_nointr_expect(ipc_sockets[0], &status, sizeof(status), &expected);
-		if (ret <= 0) {
-			ERROR("Error using IPC for the child to tell us to open LSM fd (3): %s.",
-					strerror(errno));
-			goto cleanup_error;
-		}
-
-		/* Open LSM fd and send it to child. */
-		if ((options->namespaces & CLONE_NEWNS) && (options->attach_flags & LXC_ATTACH_LSM) && init_ctx->lsm_label) {
-			int on_exec, saved_errno;
-			int labelfd = -1;
-			on_exec = options->attach_flags & LXC_ATTACH_LSM_EXEC ? 1 : 0;
-			/* Open fd for the LSM security module. */
-			labelfd = lsm_openat(procfd, attached_pid, on_exec);
-			if (labelfd < 0)
-				goto cleanup_error;
-
-			/* Send child fd of the LSM security module to write to. */
-			ret = lxc_abstract_unix_send_fd(ipc_sockets[0], labelfd, NULL, 0);
-			saved_errno = errno;
-			close(labelfd);
-			if (ret <= 0) {
-				ERROR("Intended to send file descriptor %d: %s.", labelfd, strerror(saved_errno));
-				goto cleanup_error;
-			}
-		}
-
-		if (procfd >= 0)
-			close(procfd);
 		/* now shut down communication with child, we're done */
 		shutdown(ipc_sockets[0], SHUT_RDWR);
 		close(ipc_sockets[0]);
@@ -978,8 +810,6 @@ int lxc_attach(const char* name, const char* lxcpath, lxc_attach_exec_t exec_fun
 		/* first shut down the socket, then wait for the pid,
 		 * otherwise the pid we're waiting for may never exit
 		 */
-		if (procfd >= 0)
-			close(procfd);
 		shutdown(ipc_sockets[0], SHUT_RDWR);
 		close(ipc_sockets[0]);
 		if (to_cleanup_pid)
@@ -1078,7 +908,6 @@ static int attach_child_main(void* data)
 	int expected;
 	long flags;
 	int fd;
-	int lsm_labelfd;
 	uid_t new_uid;
 	gid_t new_gid;
 
@@ -1089,7 +918,7 @@ static int attach_child_main(void* data)
 	status = -1;
 	ret = lxc_read_nointr_expect(ipc_socket, &status, sizeof(status), &expected);
 	if (ret <= 0) {
-		ERROR("Error using IPC to receive notification from initial process (0): %s.", strerror(errno));
+		ERROR("error using IPC to receive notification from initial process (0)");
 		shutdown(ipc_socket, SHUT_RDWR);
 		rexit(-1);
 	}
@@ -1156,7 +985,7 @@ static int attach_child_main(void* data)
 		new_gid = options->gid;
 
 	/* setup the control tty */
-	if (options->stdin_fd && isatty(options->stdin_fd)) {
+	if (options->stdin_fd) {
 		if (setsid() < 0) {
 			SYSERROR("unable to setsid");
 			shutdown(ipc_socket, SHUT_RDWR);
@@ -1188,7 +1017,7 @@ static int attach_child_main(void* data)
 	status = 1;
 	ret = lxc_write_nointr(ipc_socket, &status, sizeof(status));
 	if (ret != sizeof(status)) {
-		ERROR("Error using IPC to notify initial process for initialization (1): %s.", strerror(errno));
+		ERROR("error using IPC to notify initial process for initialization (1)");
 		shutdown(ipc_socket, SHUT_RDWR);
 		rexit(-1);
 	}
@@ -1200,52 +1029,32 @@ static int attach_child_main(void* data)
 	status = -1;
 	ret = lxc_read_nointr_expect(ipc_socket, &status, sizeof(status), &expected);
 	if (ret <= 0) {
-		ERROR("Error using IPC to receive message from initial process "
-		      "that it is done pre-initializing (2): %s",
-		      strerror(errno));
-		shutdown(ipc_socket, SHUT_RDWR);
-		rexit(-1);
-	}
-
-	/* Tell the (grand)parent to send us LSM label fd. */
-	status = 3;
-	ret = lxc_write_nointr(ipc_socket, &status, sizeof(status));
-	if (ret <= 0) {
-		ERROR("Error using IPC to tell parent to set up LSM labels (3): %s.", strerror(errno));
-		shutdown(ipc_socket, SHUT_RDWR);
-		rexit(-1);
-	}
-
-	if ((options->namespaces & CLONE_NEWNS) && (options->attach_flags & LXC_ATTACH_LSM) && init_ctx->lsm_label) {
-		int on_exec;
-		/* Receive fd for LSM security module. */
-		ret = lxc_abstract_unix_recv_fd(ipc_socket, &lsm_labelfd, NULL, 0);
-		if (ret <= 0) {
-			ERROR("Error using IPC for parent to tell us LSM label fd (4): %s.", strerror(errno));
-			shutdown(ipc_socket, SHUT_RDWR);
-			rexit(-1);
-		}
-
-		/* Change into our new LSM profile. */
-		on_exec = options->attach_flags & LXC_ATTACH_LSM_EXEC ? 1 : 0;
-		if (lsm_set_label_at(lsm_labelfd, on_exec, init_ctx->lsm_label) < 0) {
-			SYSERROR("Failed to set LSM label.");
-			shutdown(ipc_socket, SHUT_RDWR);
-			close(lsm_labelfd);
-			rexit(-1);
-		}
-		close(lsm_labelfd);
-	}
-
-	if (init_ctx->container && init_ctx->container->lxc_conf &&
-			lxc_seccomp_load(init_ctx->container->lxc_conf) != 0) {
-		ERROR("Loading seccomp policy");
+		ERROR("error using IPC to receive final notification from initial process (2)");
 		shutdown(ipc_socket, SHUT_RDWR);
 		rexit(-1);
 	}
 
 	shutdown(ipc_socket, SHUT_RDWR);
 	close(ipc_socket);
+
+	/* set new apparmor profile/selinux context */
+	if ((options->namespaces & CLONE_NEWNS) && (options->attach_flags & LXC_ATTACH_LSM)) {
+		int on_exec;
+
+		on_exec = options->attach_flags & LXC_ATTACH_LSM_EXEC ? 1 : 0;
+		ret = lsm_process_label_set(init_ctx->lsm_label,
+				init_ctx->container->lxc_conf, 0, on_exec);
+		if (ret < 0) {
+			rexit(-1);
+		}
+	}
+
+	if (init_ctx->container && init_ctx->container->lxc_conf &&
+			lxc_seccomp_load(init_ctx->container->lxc_conf) != 0) {
+		ERROR("Loading seccomp policy");
+		rexit(-1);
+	}
+
 	lxc_proc_put_context_info(init_ctx);
 
 	/* The following is done after the communication socket is
