@@ -39,7 +39,6 @@
 #include <sys/mount.h>
 #include <netinet/in.h>
 #include <net/if.h>
-#include <poll.h>
 
 #include "error.h"
 #include "commands.h"
@@ -122,13 +121,13 @@ static void cull_user_controllers(void);
 
 static void cgm_dbus_disconnect(void)
 {
-	if (cgroup_manager) {
-		dbus_connection_flush(cgroup_manager->connection);
-		dbus_connection_close(cgroup_manager->connection);
-		nih_free(cgroup_manager);
-	}
-	cgroup_manager = NULL;
-	cgm_unlock();
+       if (cgroup_manager) {
+	       dbus_connection_flush(cgroup_manager->connection);
+	       dbus_connection_close(cgroup_manager->connection);
+               nih_free(cgroup_manager);
+       }
+       cgroup_manager = NULL;
+       cgm_unlock();
 }
 
 #define CGMANAGER_DBUS_SOCK "unix:path=/sys/fs/cgroup/cgmanager/sock"
@@ -247,12 +246,11 @@ static void check_supports_multiple_controllers(pid_t pid)
 		}
 		if (strcmp(prevpath, colon) != 0) {
 			cgm_all_controllers_same = false;
-			break;
+			fclose(f);
+			return;
 		}
 	}
-
 	fclose(f);
-	free(line);
 }
 
 static int send_creds(int sock, int rpid, int ruid, int rgid)
@@ -348,7 +346,6 @@ static int do_chown_cgroup(const char *controller, const char *cgroup_path,
 {
 	int sv[2] = {-1, -1}, optval = 1, ret = -1;
 	char buf[1];
-	struct pollfd fds;
 
 	if (socketpair(AF_UNIX, SOCK_DGRAM, 0, sv) < 0) {
 		SYSERROR("Error creating socketpair");
@@ -372,10 +369,10 @@ static int do_chown_cgroup(const char *controller, const char *cgroup_path,
 	}
 	/* now send credentials */
 
-	fds.fd = sv[0];
-	fds.events = POLLIN;
-	fds.revents = 0;
-	if (poll(&fds, 1, -1) <= 0) {
+	fd_set rfds;
+	FD_ZERO(&rfds);
+	FD_SET(sv[0], &rfds);
+	if (select(sv[0]+1, &rfds, NULL, NULL, NULL) < 0) {
 		ERROR("Error getting go-ahead from server: %s", strerror(errno));
 		goto out;
 	}
@@ -387,10 +384,9 @@ static int do_chown_cgroup(const char *controller, const char *cgroup_path,
 		SYSERROR("%s: Error sending pid over SCM_CREDENTIAL", __func__);
 		goto out;
 	}
-	fds.fd = sv[0];
-	fds.events = POLLIN;
-	fds.revents = 0;
-	if (poll(&fds, 1, -1) <= 0) {
+	FD_ZERO(&rfds);
+	FD_SET(sv[0], &rfds);
+	if (select(sv[0]+1, &rfds, NULL, NULL, NULL) < 0) {
 		ERROR("Error getting go-ahead from server: %s", strerror(errno));
 		goto out;
 	}
@@ -402,10 +398,9 @@ static int do_chown_cgroup(const char *controller, const char *cgroup_path,
 		SYSERROR("%s: Error sending pid over SCM_CREDENTIAL", __func__);
 		goto out;
 	}
-	fds.fd = sv[0];
-	fds.events = POLLIN;
-	fds.revents = 0;
-	if (poll(&fds, 1, -1) <= 0) {
+	FD_ZERO(&rfds);
+	FD_SET(sv[0], &rfds);
+	if (select(sv[0]+1, &rfds, NULL, NULL, NULL) < 0) {
 		ERROR("Error getting go-ahead from server: %s", strerror(errno));
 		goto out;
 	}
@@ -502,9 +497,9 @@ static bool chown_cgroup(const char *cgroup_path, struct lxc_conf *conf)
 	for (i = 0; slist[i]; i++) {
 		if (!lxc_cgmanager_chmod(slist[i], cgroup_path, "", 0775))
 			return false;
-		if (!lxc_cgmanager_chmod(slist[i], cgroup_path, "tasks", 0664))
+		if (!lxc_cgmanager_chmod(slist[i], cgroup_path, "tasks", 0775))
 			return false;
-		if (!lxc_cgmanager_chmod(slist[i], cgroup_path, "cgroup.procs", 0664))
+		if (!lxc_cgmanager_chmod(slist[i], cgroup_path, "cgroup.procs", 0775))
 			return false;
 	}
 
@@ -739,6 +734,15 @@ static const char *cgm_get_cgroup(void *hdata, const char *subsystem)
 	return d->cgroup_path;
 }
 
+static const char *cgm_canonical_path(void *hdata)
+{
+	struct cgm_data *d = hdata;
+
+	if (!d || !d->cgroup_path)
+		return NULL;
+	return d->cgroup_path;
+}
+
 #if HAVE_CGMANAGER_GET_PID_CGROUP_ABS_SYNC
 static inline bool abs_cgroup_supported(void) {
 	return api_version >= CGM_SUPPORTS_GET_ABS;
@@ -766,8 +770,7 @@ static char *try_get_abs_cgroup(const char *name, const char *lxcpath,
 			NihError *nerr;
 			nerr = nih_error_get();
 			nih_free(nerr);
-		} else
-			prune_init_scope(cgroup);
+		}
 		return cgroup;
 	}
 
@@ -810,32 +813,6 @@ out:
 	return pids_len;
 }
 
-#if HAVE_CGMANAGER_LIST_CONTROLLERS
-static bool lxc_list_controllers(char ***list)
-{
-	if (!cgm_dbus_connect()) {
-		ERROR("Error connecting to cgroup manager");
-		return false;
-	}
-	if (cgmanager_list_controllers_sync(NULL, cgroup_manager, list) != 0) {
-		NihError *nerr;
-		nerr = nih_error_get();
-		ERROR("call to cgmanager_list_controllers_sync failed: %s", nerr->message);
-		nih_free(nerr);
-		cgm_dbus_disconnect();
-		return false;
-	}
-
-	cgm_dbus_disconnect();
-	return true;
-}
-#else
-static bool lxc_list_controllers(char ***list)
-{
-	return false;
-}
-#endif
-
 static inline void free_abs_cgroup(char *cgroup)
 {
 	if (!cgroup)
@@ -1111,82 +1088,11 @@ static void cull_user_controllers(void)
 	}
 }
 
-static bool in_comma_list(const char *inword, const char *cgroup_use)
-{
-	char *e;
-	size_t inlen = strlen(inword), len;
-
-	do {
-		e = strchr(cgroup_use, ',');
-		len = e ? e - cgroup_use : strlen(cgroup_use);
-		if (len == inlen && strncmp(inword, cgroup_use, len) == 0)
-			return true;
-		cgroup_use = e + 1;
-	} while (e);
-
-	return false;
-}
-
-static bool in_subsystem_list(const char *c)
-{
-	int i;
-
-	for (i = 0; i < nr_subsystems; i++) {
-		if (strcmp(c, subsystems[i]) == 0)
-			return true;
-	}
-
-	return false;
-}
-
-/*
- * If /etc/lxc/lxc.conf specifies lxc.cgroup.use = "freezer,memory",
- * then clear out any other subsystems, and make sure that freezer
- * and memory are both enabled
- */
-static bool verify_and_prune(const char *cgroup_use)
-{
-	const char *p;
-	char *e;
-	int i, j;
-
-	for (p = cgroup_use; p && *p; p = e + 1) {
-		e = strchr(p, ',');
-		if (e)
-			*e = '\0';
-
-		if (!in_subsystem_list(p)) {
-			ERROR("Controller %s required by lxc.cgroup.use but not available\n", p);
-			return false;
-		}
-
-		if (e)
-			*e = ',';
-		if (!e)
-			break;
-	}
-
-	for (i = 0; i < nr_subsystems;) {
-		if (in_comma_list(subsystems[i], cgroup_use)) {
-			i++;
-			continue;
-		}
-		free(subsystems[i]);
-		for (j = i;  j < nr_subsystems-1; j++)
-			subsystems[j] = subsystems[j+1];
-		subsystems[nr_subsystems-1] = NULL;
-		nr_subsystems--;
-	}
-
-	return true;
-}
-
 static bool collect_subsytems(void)
 {
 	char *line = NULL;
-	nih_local char **cgm_subsys_list = NULL;
 	size_t sz = 0;
-	FILE *f = NULL;
+	FILE *f;
 
 	if (subsystems) // already initialized
 		return true;
@@ -1197,20 +1103,6 @@ static bool collect_subsytems(void)
 	subsystems_inone[0] = "all";
 	subsystems_inone[1] = NULL;
 
-	if (lxc_list_controllers(&cgm_subsys_list)) {
-		while (cgm_subsys_list[nr_subsystems]) {
-			char **tmp = NIH_MUST( realloc(subsystems,
-						(nr_subsystems+2)*sizeof(char *)) );
-			tmp[nr_subsystems] = NIH_MUST(
-					strdup(cgm_subsys_list[nr_subsystems++]) );
-			subsystems = tmp;
-		}
-		if (nr_subsystems)
-			subsystems[nr_subsystems] = NULL;
-		goto collected;
-	}
-
-	INFO("cgmanager_list_controllers failed, falling back to /proc/self/cgroups");
 	f = fopen_cloexec("/proc/self/cgroup", "r");
 	if (!f) {
 		f = fopen_cloexec("/proc/1/cgroup", "r");
@@ -1250,37 +1142,18 @@ static bool collect_subsytems(void)
 		}
 	}
 	fclose(f);
-	f = NULL;
 
 	free(line);
-	line = NULL;
-
-collected:
 	if (!nr_subsystems) {
 		ERROR("No cgroup subsystems found");
 		return false;
 	}
 
-	/* make sure that cgroup.use can be and is honored */
-	const char *cgroup_use = lxc_global_config_value("lxc.cgroup.use");
-	if (!cgroup_use && errno != 0)
-		goto out_good;
-	if (cgroup_use) {
-		if (!verify_and_prune(cgroup_use)) {
-			free_subsystems();
-			return false;
-		}
-		subsystems_inone[0] = NIH_MUST( strdup(cgroup_use) );
-		cgm_all_controllers_same = false;
-	}
-
-out_good:
 	return true;
 
 out_free:
 	free(line);
-	if (f)
-		fclose(f);
+	fclose(f);
 	free_subsystems();
 	return false;
 }
@@ -1340,7 +1213,7 @@ static bool cgm_unfreeze(void *hdata)
 static bool cgm_setup_limits(void *hdata, struct lxc_list *cgroup_settings, bool do_devices)
 {
 	struct cgm_data *d = hdata;
-	struct lxc_list *iterator, *sorted_cgroup_settings, *next;
+	struct lxc_list *iterator;
 	struct lxc_cgroup *cg;
 	bool ret = false;
 
@@ -1355,12 +1228,7 @@ static bool cgm_setup_limits(void *hdata, struct lxc_list *cgroup_settings, bool
 		return false;
 	}
 
-	sorted_cgroup_settings = sort_cgroup_settings(cgroup_settings);
-	if (!sorted_cgroup_settings) {
-		return false;
-	}
-
-	lxc_list_for_each(iterator, sorted_cgroup_settings) {
+	lxc_list_for_each(iterator, cgroup_settings) {
 		char controller[100], *p;
 		cg = iterator->elem;
 		if (do_devices != !strncmp("devices", cg->subsystem, 7))
@@ -1375,14 +1243,6 @@ static bool cgm_setup_limits(void *hdata, struct lxc_list *cgroup_settings, bool
 					 d->cgroup_path, cg->subsystem, cg->value) != 0) {
 			NihError *nerr;
 			nerr = nih_error_get();
-			if (do_devices) {
-				WARN("call to cgmanager_set_value_sync failed: %s", nerr->message);
-				nih_free(nerr);
-				WARN("Error setting cgroup %s:%s limit type %s", controller,
-					d->cgroup_path, cg->subsystem);
-				continue;
-			}
-
 			ERROR("call to cgmanager_set_value_sync failed: %s", nerr->message);
 			nih_free(nerr);
 			ERROR("Error setting cgroup %s:%s limit type %s", controller,
@@ -1396,11 +1256,6 @@ static bool cgm_setup_limits(void *hdata, struct lxc_list *cgroup_settings, bool
 	ret = true;
 	INFO("cgroup limits have been setup");
 out:
-	lxc_list_for_each_safe(iterator, sorted_cgroup_settings, next) {
-		lxc_list_del(iterator);
-		free(iterator);
-	}
-	free(sorted_cgroup_settings);
 	cgm_dbus_disconnect();
 	return ret;
 }
@@ -1441,8 +1296,16 @@ static bool cgm_attach(const char *name, const char *lxcpath, pid_t pid)
 		return false;
 	}
 
+	check_supports_multiple_controllers(pid);
+
+	if (cgm_all_controllers_same)
+		slist = subsystems_inone;
+
 	for (i = 0; slist[i]; i++) {
-		cgroup = try_get_abs_cgroup(name, lxcpath, slist[i]);
+		if (slist == subsystems_inone)
+			cgroup = try_get_abs_cgroup(name, lxcpath, subsystems[0]);
+		else
+			cgroup = try_get_abs_cgroup(name, lxcpath, slist[i]);
 		if (!cgroup) {
 			ERROR("Failed to get cgroup for controller %s", slist[i]);
 			cgm_dbus_disconnect();
@@ -1477,7 +1340,7 @@ static bool cgm_bind_dir(const char *root, const char *dirname)
 	}
 
 	/* mount a tmpfs there so we can create subdirs */
-	if (safe_mount("cgroup", cgpath, "tmpfs", 0, "size=10000,mode=755", root)) {
+	if (mount("cgroup", cgpath, "tmpfs", 0, "size=10000,mode=755")) {
 		SYSERROR("Failed to mount tmpfs at %s", cgpath);
 		return false;
 	}
@@ -1488,7 +1351,7 @@ static bool cgm_bind_dir(const char *root, const char *dirname)
 		return false;
 	}
 
-	if (safe_mount(dirname, cgpath, "none", MS_BIND, 0, root)) {
+	if (mount(dirname, cgpath, "none", MS_BIND, 0)) {
 		SYSERROR("Failed to bind mount %s to %s", dirname, cgpath);
 		return false;
 	}
@@ -1522,6 +1385,7 @@ static struct cgroup_ops cgmanager_ops = {
 	.enter = cgm_enter,
 	.create_legacy = NULL,
 	.get_cgroup = cgm_get_cgroup,
+	.canonical_path = cgm_canonical_path,
 	.get = cgm_get,
 	.set = cgm_set,
 	.unfreeze = cgm_unfreeze,
