@@ -21,6 +21,9 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
+#include <unistd.h>
+#include <sys/types.h>
+
 #include "cgroup.h"
 #include "conf.h"
 #include "log.h"
@@ -106,6 +109,19 @@ const char *cgroup_get_cgroup(struct lxc_handler *handler, const char *subsystem
 	return NULL;
 }
 
+const char *cgroup_canonical_path(struct lxc_handler *handler)
+{
+	if (geteuid()) {
+		WARN("cgroup_canonical_path only makes sense for privileged containers.\n");
+		return NULL;
+	}
+
+	if (ops)
+		return ops->canonical_path(handler->cgroup_data);
+
+	return NULL;
+}
+
 bool cgroup_unfreeze(struct lxc_handler *handler)
 {
 	if (ops)
@@ -173,22 +189,3 @@ void cgroup_disconnect(void)
 	if (ops && ops->disconnect)
 		ops->disconnect();
 }
-
-#define INIT_SCOPE "/init.scope"
-void prune_init_scope(char *cg)
-{
-	char *point;
-
-	if (!cg)
-		return;
-
-	point = cg + strlen(cg) - strlen(INIT_SCOPE);
-	if (point < cg)
-		return;
-	if (strcmp(point, INIT_SCOPE) == 0) {
-		if (point == cg)
-			*(point+1) = '\0';
-		else
-			*point = '\0';
-	}
-}
