@@ -74,19 +74,14 @@
 
 lxc_log_define(lxc_commands, lxc);
 
-static int fill_sock_name(char *path, int len, const char *lxcname,
+static int fill_sock_name(char *path, int len, const char *name,
 			  const char *lxcpath, const char *hashed_sock_name)
 {
-	const char *name;
 	char *tmppath;
 	size_t tmplen;
 	uint64_t hash;
 	int ret;
 
-	name = lxcname;
-	if (!name)
-		name = "";
-
 	if (hashed_sock_name != NULL) {
 		ret = snprintf(path, len, "lxc/%s/command", hashed_sock_name);
 		if (ret < 0 || ret >= len) {
@@ -173,8 +168,8 @@ static int lxc_cmd_rsp_recv(int sock, struct lxc_cmd_rr *cmd)
 
 	ret = lxc_abstract_unix_recv_fd(sock, &rspfd, rsp, sizeof(*rsp));
 	if (ret < 0) {
-		WARN("Command %s failed to receive response: %s.",
-		     lxc_cmd_str(cmd->req.cmd), strerror(errno));
+		WARN("command %s failed to receive response",
+		     lxc_cmd_str(cmd->req.cmd));
 		return -1;
 	}
 
@@ -189,7 +184,7 @@ static int lxc_cmd_rsp_recv(int sock, struct lxc_cmd_rr *cmd)
 
 		rspdata = malloc(sizeof(*rspdata));
 		if (!rspdata) {
-			ERROR("Command %s couldn't allocate response buffer.",
+			ERROR("command %s couldn't allocate response buffer",
 			      lxc_cmd_str(cmd->req.cmd));
 			return -1;
 		}
@@ -198,13 +193,10 @@ static int lxc_cmd_rsp_recv(int sock, struct lxc_cmd_rr *cmd)
 		rsp->data = rspdata;
 	}
 
-	if (rsp->datalen == 0) {
-		DEBUG("command %s response data length is 0",
-		      lxc_cmd_str(cmd->req.cmd));
+	if (rsp->datalen == 0)
 		return ret;
-	}
 	if (rsp->datalen > LXC_CMD_DATA_MAX) {
-		ERROR("Command %s response data %d too long.",
+		ERROR("command %s response data %d too long",
 		      lxc_cmd_str(cmd->req.cmd), rsp->datalen);
 		errno = EFBIG;
 		return -1;
@@ -212,14 +204,14 @@ static int lxc_cmd_rsp_recv(int sock, struct lxc_cmd_rr *cmd)
 
 	rsp->data = malloc(rsp->datalen);
 	if (!rsp->data) {
-		ERROR("Command %s was unable to allocate response buffer.",
+		ERROR("command %s unable to allocate response buffer",
 		      lxc_cmd_str(cmd->req.cmd));
 		return -1;
 	}
 	ret = recv(sock, rsp->data, rsp->datalen, 0);
 	if (ret != rsp->datalen) {
-		ERROR("Command %s failed to receive response data: %s.",
-		      lxc_cmd_str(cmd->req.cmd), strerror(errno));
+		ERROR("command %s failed to receive response data",
+		      lxc_cmd_str(cmd->req.cmd));
 		if (ret >= 0)
 			ret = -1;
 	}
@@ -241,7 +233,7 @@ static int lxc_cmd_rsp_send(int fd, struct lxc_cmd_rsp *rsp)
 
 	ret = send(fd, rsp, sizeof(*rsp), 0);
 	if (ret != sizeof(*rsp)) {
-		ERROR("Failed to send command response %d: %s.", ret,
+		ERROR("failed to send command response %d %s", ret,
 		      strerror(errno));
 		return -1;
 	}
@@ -249,8 +241,8 @@ static int lxc_cmd_rsp_send(int fd, struct lxc_cmd_rsp *rsp)
 	if (rsp->datalen > 0) {
 		ret = send(fd, rsp->data, rsp->datalen, 0);
 		if (ret != rsp->datalen) {
-			WARN("Failed to send command response data %d: %s.",
-			     ret, strerror(errno));
+			WARN("failed to send command response data %d %s", ret,
+			      strerror(errno));
 			return -1;
 		}
 	}
@@ -282,17 +274,12 @@ static int lxc_cmd(const char *name, struct lxc_cmd_rr *cmd, int *stopped,
 	int sock, ret = -1;
 	char path[sizeof(((struct sockaddr_un *)0)->sun_path)] = { 0 };
 	char *offset = &path[1];
-	size_t len;
+	int len;
 	int stay_connected = cmd->req.cmd == LXC_CMD_CONSOLE;
 
 	*stopped = 0;
 
-	/* -2 here because this is an abstract unix socket so it needs a
-	 * leading \0, and we null terminate, so it needs a trailing \0.
-	 * Although null termination isn't required by the API, we do it anyway
-	 * because we print the sockname out sometimes.
-	 */
-	len = sizeof(path)-2;
+	len = sizeof(path)-1;
 	if (fill_sock_name(offset, len, name, lxcpath, hashed_sock_name))
 		return -1;
 
@@ -301,7 +288,7 @@ static int lxc_cmd(const char *name, struct lxc_cmd_rr *cmd, int *stopped,
 		if (errno == ECONNREFUSED)
 			*stopped = 1;
 		else
-			SYSERROR("Command %s failed to connect to \"@%s\".",
+			SYSERROR("command %s failed to connect to '@%s'",
 				 lxc_cmd_str(cmd->req.cmd), offset);
 		return -1;
 	}
@@ -310,7 +297,7 @@ static int lxc_cmd(const char *name, struct lxc_cmd_rr *cmd, int *stopped,
 	if (ret != sizeof(cmd->req)) {
 		if (errno == EPIPE)
 			goto epipe;
-		SYSERROR("Command %s failed to send req to \"@%s\" %d.",
+		SYSERROR("command %s failed to send req to '@%s' %d",
 			 lxc_cmd_str(cmd->req.cmd), offset, ret);
 		if (ret >=0)
 			ret = -1;
@@ -322,7 +309,7 @@ static int lxc_cmd(const char *name, struct lxc_cmd_rr *cmd, int *stopped,
 		if (ret != cmd->req.datalen) {
 			if (errno == EPIPE)
 				goto epipe;
-			SYSERROR("Command %s failed to send request data to \"@%s\" %d.",
+			SYSERROR("command %s failed to send request data to '@%s' %d",
 				 lxc_cmd_str(cmd->req.cmd), offset, ret);
 			if (ret >=0)
 				ret = -1;
@@ -466,13 +453,14 @@ char *lxc_cmd_get_cgroup_path(const char *name, const char *lxcpath,
 		return NULL;
 
 	if (!ret) {
-		WARN("Container \"%s\" has stopped before sending its state.", name);
+		WARN("'%s' has stopped before sending its state", name);
 		return NULL;
 	}
 
 	if (cmd.rsp.ret < 0 || cmd.rsp.datalen < 0) {
-		ERROR("Command %s failed for container \"%s\": %s.",
-		      lxc_cmd_str(cmd.req.cmd), name, strerror(-cmd.rsp.ret));
+		ERROR("command %s failed for '%s': %s",
+		      lxc_cmd_str(cmd.req.cmd), name,
+		      strerror(-cmd.rsp.ret));
 		return NULL;
 	}
 
@@ -578,11 +566,11 @@ lxc_state_t lxc_cmd_get_state(const char *name, const char *lxcpath)
 		return -1;
 
 	if (!ret) {
-		WARN("Container \"%s\" has stopped before sending its state.", name);
+		WARN("'%s' has stopped before sending its state", name);
 		return -1;
 	}
 
-	DEBUG("Container \"%s\" is in \"%s\" state.", name,
+	DEBUG("'%s' is in '%s' state", name,
 	      lxc_state2str(PTR_TO_INT(cmd.rsp.data)));
 	return PTR_TO_INT(cmd.rsp.data);
 }
@@ -614,7 +602,7 @@ int lxc_cmd_stop(const char *name, const char *lxcpath)
 	ret = lxc_cmd(name, &cmd, &stopped, lxcpath, NULL);
 	if (ret < 0) {
 		if (stopped) {
-			INFO("Container \"%s\" is already stopped.", name);
+			INFO("'%s' is already stopped", name);
 			return 0;
 		}
 		return -1;
@@ -624,12 +612,11 @@ int lxc_cmd_stop(const char *name, const char *lxcpath)
 	 * closed
 	 */
 	if (ret > 0) {
-		ERROR("Failed to stop container \"%s\": %s.", name,
-		      strerror(-cmd.rsp.ret));
+		ERROR("failed to stop '%s': %s", name, strerror(-cmd.rsp.ret));
 		return -1;
 	}
 
-	INFO("Container \"%s\" has stopped.", name);
+	INFO("'%s' has stopped", name);
 	return 0;
 }
 
@@ -651,7 +638,7 @@ static int lxc_cmd_stop_callback(int fd, struct lxc_cmd_req *req,
 		 */
 		if (cgroup_unfreeze(handler))
 			return 0;
-		ERROR("Failed to unfreeze container \"%s\".", handler->name);
+		ERROR("Failed to unfreeze %s:%s", handler->lxcpath, handler->name);
 		rsp.ret = -1;
 	}
 
@@ -713,27 +700,27 @@ int lxc_cmd_console(const char *name, int *ttynum, int *fd, const char *lxcpath)
 		return ret;
 
 	if (cmd.rsp.ret < 0) {
-		ERROR("Console access denied: %s.", strerror(-cmd.rsp.ret));
+		ERROR("console access denied: %s", strerror(-cmd.rsp.ret));
 		ret = -1;
 		goto out;
 	}
 
 	if (ret == 0) {
-		ERROR("Console %d invalid, busy or all consoles busy.", *ttynum);
+		ERROR("console %d invalid,busy or all consoles busy", *ttynum);
 		ret = -1;
 		goto out;
 	}
 
 	rspdata = cmd.rsp.data;
 	if (rspdata->masterfd < 0) {
-		ERROR("Unable to allocate fd for tty %d.", rspdata->ttynum);
+		ERROR("unable to allocate fd for tty %d", rspdata->ttynum);
 		goto out;
 	}
 
 	ret = cmd.rsp.ret;	/* sock fd */
 	*fd = rspdata->masterfd;
 	*ttynum = rspdata->ttynum;
-	INFO("tty %d allocated fd %d sock %d.", rspdata->ttynum, *fd, ret);
+	INFO("tty %d allocated fd %d sock %d", rspdata->ttynum, *fd, ret);
 out:
 	free(cmd.rsp.data);
 	return ret;
@@ -753,7 +740,7 @@ static int lxc_cmd_console_callback(int fd, struct lxc_cmd_req *req,
 	memset(&rsp, 0, sizeof(rsp));
 	rsp.data = INT_TO_PTR(ttynum);
 	if (lxc_abstract_unix_send_fd(fd, masterfd, &rsp, sizeof(rsp)) < 0) {
-		ERROR("Failed to send tty to client.");
+		ERROR("failed to send tty to client");
 		lxc_console_free(handler->conf, fd);
 		goto out_close;
 	}
@@ -862,7 +849,7 @@ static int lxc_cmd_process(int fd, struct lxc_cmd_req *req,
 	};
 
 	if (req->cmd >= LXC_CMD_MAX) {
-		ERROR("Undefined command id %d received.", req->cmd);
+		ERROR("bad cmd %d received", req->cmd);
 		return -1;
 	}
 	return cb[req->cmd](fd, req, handler);
@@ -893,23 +880,23 @@ static int lxc_cmd_handler(int fd, uint32_t events, void *data,
 	}
 
 	if (ret < 0) {
-		SYSERROR("Failed to receive data on command socket.");
+		SYSERROR("failed to receive data on command socket");
 		goto out_close;
 	}
 
 	if (!ret) {
-		DEBUG("Peer has disconnected.");
+		DEBUG("peer has disconnected");
 		goto out_close;
 	}
 
 	if (ret != sizeof(req)) {
-		WARN("Failed to receive full command request. Ignoring request.");
+		WARN("partial request, ignored");
 		ret = -1;
 		goto out_close;
 	}
 
 	if (req.datalen > LXC_CMD_DATA_MAX) {
-		ERROR("Received command data length %d is too large.", req.datalen);
+		ERROR("cmd data length %d too large", req.datalen);
 		ret = -1;
 		goto out_close;
 	}
@@ -920,7 +907,7 @@ static int lxc_cmd_handler(int fd, uint32_t events, void *data,
 		reqdata = alloca(req.datalen);
 		ret = recv(fd, reqdata, req.datalen, 0);
 		if (ret != req.datalen) {
-			WARN("Failed to receive full command request. Ignoring request.");
+			WARN("partial request, ignored");
 			ret = -1;
 			goto out_close;
 		}
@@ -948,24 +935,24 @@ static int lxc_cmd_accept(int fd, uint32_t events, void *data,
 
 	connection = accept(fd, NULL, 0);
 	if (connection < 0) {
-		SYSERROR("Failed to accept connection to run command.");
+		SYSERROR("failed to accept connection");
 		return -1;
 	}
 
 	if (fcntl(connection, F_SETFD, FD_CLOEXEC)) {
-		SYSERROR("Failed to set close-on-exec on incoming command connection.");
+		SYSERROR("failed to set close-on-exec on incoming connection");
 		goto out_close;
 	}
 
 	if (setsockopt(connection, SOL_SOCKET,
 		       SO_PASSCRED, &opt, sizeof(opt))) {
-		SYSERROR("Failed to enable necessary credentials on command socket.");
+		SYSERROR("failed to enable credential on socket");
 		goto out_close;
 	}
 
 	ret = lxc_mainloop_add_handler(descr, connection, lxc_cmd_handler, data);
 	if (ret) {
-		ERROR("Failed to add command handler.");
+		ERROR("failed to add handler");
 		goto out_close;
 	}
 
@@ -985,26 +972,23 @@ int lxc_cmd_init(const char *name, struct lxc_handler *handler,
 	char *offset = &path[1];
 	int len;
 
-	/* -2 here because this is an abstract unix socket so it needs a
-	 * leading \0, and we null terminate, so it needs a trailing \0.
-	 * Although null termination isn't required by the API, we do it anyway
-	 * because we print the sockname out sometimes.
-	 */
-	len = sizeof(path) - 2;
+	len = sizeof(path)-1;
 	if (fill_sock_name(offset, len, name, lxcpath, NULL))
 		return -1;
 
 	fd = lxc_abstract_unix_open(path, SOCK_STREAM, 0);
 	if (fd < 0) {
-		ERROR("Failed to create the command service point %s: %s.",
-		      offset, strerror(errno));
-		if (errno == EADDRINUSE)
-			ERROR("Container \"%s\" appears to be already running!", name);
+		ERROR("failed (%d) to create the command service point %s", errno, offset);
+		if (errno == EADDRINUSE) {
+			ERROR("##");
+			ERROR("# The container appears to be already running!");
+			ERROR("##");
+		}
 		return -1;
 	}
 
 	if (fcntl(fd, F_SETFD, FD_CLOEXEC)) {
-		SYSERROR("Failed to set FD_CLOEXEC on signal file descriptor.");
+		SYSERROR("failed to set sigfd to close-on-exec");
 		close(fd);
 		return -1;
 	}
@@ -1021,7 +1005,7 @@ int lxc_cmd_mainloop_add(const char *name,
 
 	ret = lxc_mainloop_add_handler(descr, fd, lxc_cmd_accept, handler);
 	if (ret) {
-		ERROR("Failed to add handler for command socket.");
+		ERROR("failed to add handler for command socket");
 		close(fd);
 	}
 
