@@ -20,7 +20,6 @@
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
-#include <assert.h>
 #include <stdio.h>
 #include <errno.h>
 #include <limits.h>
@@ -101,13 +100,10 @@ static int log_append_logfile(const struct lxc_log_appender *appender,
 		     event->locinfo->file, event->locinfo->func,
 		     event->locinfo->line);
 
-	if (n < 0)
-		return n;
+	n += vsnprintf(buffer + n, sizeof(buffer) - n, event->fmt,
+		       *event->vap);
 
-	if (n < sizeof(buffer) - 1)
-		n += vsnprintf(buffer + n, sizeof(buffer) - n, event->fmt,
-			       *event->vap);
-	else {
+	if (n >= sizeof(buffer) - 1) {
 		WARN("truncated next event from %d to %zd bytes", n,
 		     sizeof(buffer));
 		n = sizeof(buffer) - 1;
@@ -254,16 +250,6 @@ static char *build_log_path(const char *name, const char *lxcpath)
 	return p;
 }
 
-extern void lxc_log_close(void)
-{
-	if (lxc_log_fd == -1)
-		return;
-	close(lxc_log_fd);
-	lxc_log_fd = -1;
-	free(log_fname);
-	log_fname = NULL;
-}
-
 /*
  * This can be called:
  *   1. when a program calls lxc_log_init with no logfile parameter (in which
@@ -276,12 +262,11 @@ static int __lxc_log_set_file(const char *fname, int create_dirs)
 {
 	if (lxc_log_fd != -1) {
 		// we are overriding the default.
-		lxc_log_close();
+		close(lxc_log_fd);
+		free(log_fname);
 	}
 
-	assert(fname != NULL);
-
-	if (strlen(fname) == 0) {
+	if (!fname || strlen(fname) == 0) {
 		log_fname = NULL;
 		return 0;
 	}
@@ -351,11 +336,6 @@ extern int lxc_log_init(const char *name, const char *file,
 			return 0;
 		ret = __lxc_log_set_file(file, 1);
 	} else {
-
-		/* For now, unprivileged containers have to set -l to get logging */
-		if (geteuid())
-			return 0;
-
 		/* if no name was specified, there nothing to do */
 		if (!name)
 			return 0;
@@ -366,7 +346,7 @@ extern int lxc_log_init(const char *name, const char *file,
 			lxcpath = LOGPATH;
 
 		/* try LOGPATH if lxcpath is the default for the privileged containers */
-		if (!geteuid() && strcmp(LXCPATH, lxcpath) == 0)
+		if (!geteuid() && strcmp(lxcpath, lxc_global_config_value("lxc.lxcpath")) == 0)
 			ret = _lxc_log_set_file(name, NULL, 0);
 
 		/* try in lxcpath */
@@ -390,6 +370,16 @@ extern int lxc_log_init(const char *name, const char *file,
 	return ret;
 }
 
+extern void lxc_log_close(void)
+{
+	if (lxc_log_fd == -1)
+		return;
+	close(lxc_log_fd);
+	lxc_log_fd = -1;
+	free(log_fname);
+	log_fname = NULL;
+}
+
 /*
  * This is called when we read a lxc.loglevel entry in a lxc.conf file.  This
  * happens after processing command line arguments, which override the .conf
