@@ -48,6 +48,7 @@
 #include "nl.h"
 #include "network.h"
 #include "conf.h"
+#include "utils.h"
 
 #if HAVE_IFADDRS_H
 #include <ifaddrs.h>
@@ -108,8 +109,6 @@ int lxc_netdev_move_by_index(int ifindex, pid_t pid, const char* ifname)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi)
-		goto out;
 	ifi->ifi_family = AF_UNSPEC;
 	ifi->ifi_index = ifindex;
 
@@ -128,9 +127,115 @@ out:
 	return err;
 }
 
+/*
+ * If we are asked to move a wireless interface, then
+ * we must actually move its phyN device.  Detect
+ * that condition and return the physname here.  The
+ * physname will be passed to lxc_netdev_move_wlan()
+ * which will free it when done
+ */
+#define PHYSNAME "/sys/class/net/%s/phy80211/name"
+static char * is_wlan(const char *ifname)
+{
+	char *path, *physname = NULL;
+	size_t len = strlen(ifname) + strlen(PHYSNAME) - 1;
+	struct stat sb;
+	long physlen;
+	FILE *f;
+	int ret, i;
+
+	path = alloca(len+1);
+	ret = snprintf(path, len, PHYSNAME, ifname);
+	if (ret < 0 || ret >= len)
+		goto bad;
+	ret = stat(path, &sb);
+	if (ret)
+		goto bad;
+	if (!(f = fopen(path, "r")))
+		goto bad;
+	// feh - sb.st_size is always 4096
+	fseek(f, 0, SEEK_END);
+	physlen = ftell(f);
+	fseek(f, 0, SEEK_SET);
+	physname = malloc(physlen+1);
+	if (!physname)
+		goto bad;
+	memset(physname, 0, physlen+1);
+	ret = fread(physname, 1, physlen, f);
+	fclose(f);
+	if (ret < 0)
+		goto bad;
+
+	for (i = 0;  i < physlen; i++) {
+		if (physname[i] == '\n')
+			physname[i] = '\0';
+		if (physname[i] == '\0')
+			break;
+	}
+
+	return physname;
+
+bad:
+	free(physname);
+	return NULL;
+}
+
+static int
+lxc_netdev_rename_by_name_in_netns(pid_t pid, const char *old, const char *new)
+{
+	pid_t fpid = fork();
+
+	if (fpid < 0)
+		return -1;
+	if (fpid != 0)
+		return wait_for_pid(fpid);
+	if (!switch_to_ns(pid, "net"))
+		return -1;
+	exit(lxc_netdev_rename_by_name(old, new));
+}
+
+static int
+lxc_netdev_move_wlan(char *physname, const char *ifname, pid_t pid, const char* newname)
+{
+	int err = -1;
+	pid_t fpid;
+	char *cmd;
+
+	/* Move phyN into the container.  TODO - do this using netlink.
+	 * However, IIUC this involves a bit more complicated work to
+	 * talk to the 80211 module, so for now just call out to iw
+	 */
+	cmd = on_path("iw", NULL);
+	if (!cmd)
+		goto out1;
+	free(cmd);
+
+	fpid = fork();
+	if (fpid < 0)
+		goto out1;
+	if (fpid == 0) {
+		char pidstr[30];
+		sprintf(pidstr, "%d", pid);
+		if (execlp("iw", "iw", "phy", physname, "set", "netns", pidstr, NULL))
+			exit(1);
+		exit(0); // notreached
+	}
+	if (wait_for_pid(fpid))
+		goto out1;
+
+	err = 0;
+	if (newname)
+		err = lxc_netdev_rename_by_name_in_netns(pid, ifname, newname);
+
+out1:
+	free(physname);
+	return err;
+}
+
 int lxc_netdev_move_by_name(const char *ifname, pid_t pid, const char* newname)
 {
 	int index;
+	char *physname;
 
 	if (!ifname)
 		return -EINVAL;
@@ -139,6 +244,9 @@ int lxc_netdev_move_by_name(const char *ifname, pid_t pid, const char* newname)
 	if (!index)
 		return -EINVAL;
 
+	if ((physname = is_wlan(ifname)))
+		return lxc_netdev_move_wlan(physname, ifname, pid, newname);
+
 	return lxc_netdev_move_by_index(index, pid, newname);
 }
 
@@ -166,8 +274,6 @@ int lxc_netdev_delete_by_index(int ifindex)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_DELLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi)
-		goto out;
 	ifi->ifi_family = AF_UNSPEC;
 	ifi->ifi_index = ifindex;
 
@@ -218,8 +324,6 @@ int lxc_netdev_rename_by_index(int ifindex, const char *newname)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi)
-		goto out;
 	ifi->ifi_family = AF_UNSPEC;
 	ifi->ifi_index = ifindex;
 
@@ -283,10 +387,6 @@ int netdev_set_flag(const char *name, int flag)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi) {
-		err = -ENOMEM;
-		goto out;
-	}
 	ifi->ifi_family = AF_UNSPEC;
 	ifi->ifi_index = index;
 	ifi->ifi_change |= IFF_UP;
@@ -337,10 +437,6 @@ int netdev_get_flag(const char* name, int *flag)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_GETLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi) {
-		err = -ENOMEM;
-		goto out;
-	}
 	ifi->ifi_family = AF_UNSPEC;
 	ifi->ifi_index = index;
 
@@ -415,8 +511,6 @@ int netdev_get_mtu(int ifindex)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_GETLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi)
-		goto out;
 	ifi->ifi_family = AF_UNSPEC;
 
 	/* Send the request for addresses, which returns all addresses
@@ -528,10 +622,6 @@ int lxc_netdev_set_mtu(const char *name, int mtu)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi) {
-		err = -ENOMEM;
-		goto out;
-	}
 	ifi->ifi_family = AF_UNSPEC;
 	ifi->ifi_index = index;
 
@@ -591,8 +681,6 @@ int lxc_veth_create(const char *name1, const char *name2)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi)
-		goto out;
 	ifi->ifi_family = AF_UNSPEC;
 
 	err = -EINVAL;
@@ -612,10 +700,8 @@ int lxc_veth_create(const char *name1, const char *name2)
 		goto out;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi) {
-		err = -ENOMEM;
+	if (!ifi)
 		goto out;
-	}
 
 	if (nla_put_string(nlmsg, IFLA_IFNAME, name2))
 		goto out;
@@ -678,10 +764,6 @@ int lxc_vlan_create(const char *master, const char *name, unsigned short vlanid)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi) {
-		err = -ENOMEM;
-		goto err1;
-	}
 	ifi->ifi_family = AF_UNSPEC;
 
 	nest = nla_begin_nested(nlmsg, IFLA_LINKINFO);
@@ -758,10 +840,6 @@ int lxc_macvlan_create(const char *master, const char *name, int mode)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWLINK;
 
 	ifi = nlmsg_reserve(nlmsg, sizeof(struct ifinfomsg));
-	if (!ifi) {
-		err = -ENOMEM;
-		goto out;
-	}
 	ifi->ifi_family = AF_UNSPEC;
 
 	nest = nla_begin_nested(nlmsg, IFLA_LINKINFO);
@@ -943,8 +1021,6 @@ static int ip_addr_add(int family, int ifindex,
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWADDR;
 
 	ifa = nlmsg_reserve(nlmsg, sizeof(struct ifaddrmsg));
-	if (!ifa) 
-		goto out;
 	ifa->ifa_prefixlen = prefix;
 	ifa->ifa_index = ifindex;
 	ifa->ifa_family = family;
@@ -1066,8 +1142,6 @@ static int ip_addr_get(int family, int ifindex, void **res)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_GETADDR;
 
 	ifa = nlmsg_reserve(nlmsg, sizeof(struct ifaddrmsg));
-	if (!ifa)
-		goto out;
 	ifa->ifa_family = family;
 
 	/* Send the request for addresses, which returns all addresses
@@ -1182,8 +1256,6 @@ static int ip_gateway_add(int family, int ifindex, void *gw)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWROUTE;
 
 	rt = nlmsg_reserve(nlmsg, sizeof(struct rtmsg));
-	if (!rt)
-		goto out;
 	rt->rtm_family = family;
 	rt->rtm_table = RT_TABLE_MAIN;
 	rt->rtm_scope = RT_SCOPE_UNIVERSE;
@@ -1248,8 +1320,6 @@ static int ip_route_dest_add(int family, int ifindex, void *dest)
 	nlmsg->nlmsghdr->nlmsg_type = RTM_NEWROUTE;
 
 	rt = nlmsg_reserve(nlmsg, sizeof(struct rtmsg));
-	if (!rt)
-		goto out;
 	rt->rtm_family = family;
 	rt->rtm_table = RT_TABLE_MAIN;
 	rt->rtm_scope = RT_SCOPE_LINK;
@@ -1280,6 +1350,39 @@ int lxc_ipv6_dest_add(int ifindex, struct in6_addr *dest)
 	return ip_route_dest_add(AF_INET6, ifindex, dest);
 }
 
+static bool is_ovs_bridge(const char *bridge)
+{
+	char brdirname[22 + IFNAMSIZ + 1] = {0};
+	struct stat sb;
+
+	snprintf(brdirname, 22 +IFNAMSIZ + 1, "/sys/class/net/%s/bridge", bridge);
+	if (stat(brdirname, &sb) == -1 && errno == ENOENT)
+		return true;
+	return false;
+}
+
+static int attach_to_ovs_bridge(const char *bridge, const char *nic)
+{
+	pid_t pid;
+	char *cmd;
+
+	cmd = on_path("ovs-vsctl", NULL);
+	if (!cmd)
+		return -1;
+	free(cmd);
+
+	pid = fork();
+	if (pid < 0)
+		return -1;
+	if (pid > 0)
+		return wait_for_pid(pid);
+
+	if (execlp("ovs-vsctl", "ovs-vsctl", "add-port", bridge, nic, NULL))
+		exit(1);
+	// not reached
+	exit(1);
+}
+
 /*
  * There is a lxc_bridge_attach, but no need of a bridge detach
  * as automatically done by kernel when a netdev is deleted.
@@ -1296,6 +1399,9 @@ int lxc_bridge_attach(const char *bridge, const char *ifname)
 	if (!index)
 		return -EINVAL;
 
+	if (is_ovs_bridge(bridge))
+		return attach_to_ovs_bridge(bridge, ifname);
+
 	fd = socket(AF_INET, SOCK_STREAM, 0);
 	if (fd < 0)
 		return -errno;
