@@ -29,7 +29,6 @@
 #include <stddef.h>
 #include <string.h>
 #include <sys/types.h>
-#include <sys/vfs.h>
 #include <sys/stat.h>
 #include <sys/mman.h>
 #include <sys/param.h>
@@ -40,32 +39,43 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <assert.h>
+#include <sys/prctl.h>
 
 #include "utils.h"
 #include "log.h"
 #include "lxclock.h"
+#include "namespace.h"
 
-#ifndef O_PATH
-#define O_PATH      010000000
+#ifndef PR_SET_MM
+#define PR_SET_MM 35
 #endif
 
-#ifndef O_NOFOLLOW
-#define O_NOFOLLOW  00400000
+#ifndef PR_SET_MM_ARG_START
+#define PR_SET_MM_ARG_START 8
+#endif
+
+#ifndef PR_SET_MM_ARG_END
+#define PR_SET_MM_ARG_END 9
+#endif
+
+#ifndef PR_SET_MM_ENV_START
+#define PR_SET_MM_ENV_START 10
+#endif
+
+#ifndef PR_SET_MM_ENV_END
+#define PR_SET_MM_ENV_END 11
 #endif
 
 lxc_log_define(lxc_utils, lxc);
 
-/*
- * if path is btrfs, tries to remove it and any subvolumes beneath it
- */
-extern bool btrfs_try_remove_subvol(const char *path);
-
-static int _recursive_rmdir(char *dirname, dev_t pdev, bool onedev)
+static int _recursive_rmdir_onedev(char *dirname, dev_t pdev,
+				   const char *exclude, int level)
 {
-	struct dirent *direntp;
+	struct dirent dirent, *direntp;
 	DIR *dir;
 	int ret, failed=0;
 	char pathname[MAXPATHLEN];
+	bool hadexclude = false;
 
 	dir = opendir(dirname);
 	if (!dir) {
@@ -73,7 +83,7 @@ static int _recursive_rmdir(char *dirname, dev_t pdev, bool onedev)
 		return -1;
 	}
 
-	while ((direntp = readdir(dir))) {
+	while (!readdir_r(dir, &dirent, &direntp)) {
 		struct stat mystat;
 		int rc;
 
@@ -91,33 +101,52 @@ static int _recursive_rmdir(char *dirname, dev_t pdev, bool onedev)
 			continue;
 		}
 
+		if (!level && exclude && !strcmp(direntp->d_name, exclude)) {
+			ret = rmdir(pathname);
+			if (ret < 0) {
+				switch(errno) {
+				case ENOTEMPTY:
+					INFO("Not deleting snapshots");
+					hadexclude = true;
+					break;
+				case ENOTDIR:
+					ret = unlink(pathname);
+					if (ret)
+						INFO("%s: failed to remove %s", __func__, pathname);
+					break;
+				default:
+					SYSERROR("%s: failed to rmdir %s", __func__, pathname);
+					failed = 1;
+					break;
+				}
+			}
+			continue;
+		}
+
 		ret = lstat(pathname, &mystat);
 		if (ret) {
 			ERROR("%s: failed to stat %s", __func__, pathname);
-			failed = 1;
+			failed=1;
 			continue;
 		}
-		if (onedev && mystat.st_dev != pdev) {
-			/* TODO should we be checking /proc/self/mountinfo for
-			 * pathname and not doing this if found? */
-			if (btrfs_try_remove_subvol(pathname))
-				INFO("Removed btrfs subvolume at %s\n", pathname);
+		if (mystat.st_dev != pdev)
 			continue;
-		}
 		if (S_ISDIR(mystat.st_mode)) {
-			if (_recursive_rmdir(pathname, pdev, onedev) < 0)
+			if (_recursive_rmdir_onedev(pathname, pdev, exclude, level+1) < 0)
 				failed=1;
 		} else {
 			if (unlink(pathname) < 0) {
-				SYSERROR("%s: failed to delete %s", __func__, pathname);
+				ERROR("%s: failed to delete %s", __func__, pathname);
 				failed=1;
 			}
 		}
 	}
 
-	if (rmdir(dirname) < 0 && !btrfs_try_remove_subvol(dirname)) {
-		ERROR("%s: failed to delete %s", __func__, dirname);
-		failed=1;
+	if (rmdir(dirname) < 0) {
+		if (!hadexclude) {
+			ERROR("%s: failed to delete %s", __func__, dirname);
+			failed=1;
+		}
 	}
 
 	ret = closedir(dir);
@@ -129,43 +158,54 @@ static int _recursive_rmdir(char *dirname, dev_t pdev, bool onedev)
 	return failed ? -1 : 0;
 }
 
-/* we have two different magic values for overlayfs, yay */
-#define OVERLAYFS_SUPER_MAGIC 0x794c764f
-#define OVERLAY_SUPER_MAGIC 0x794c7630
-/*
- * In overlayfs, st_dev is unreliable.  so on overlayfs we don't do
- * the lxc_rmdir_onedev()
- */
-static bool is_native_overlayfs(const char *path)
-{
-	struct statfs sb;
-
-	if (statfs(path, &sb) < 0)
-		return false;
-	if (sb.f_type == OVERLAYFS_SUPER_MAGIC ||
-			sb.f_type == OVERLAY_SUPER_MAGIC)
-		return true;
-	return false;
-}
-
 /* returns 0 on success, -1 if there were any failures */
-extern int lxc_rmdir_onedev(char *path)
+extern int lxc_rmdir_onedev(char *path, const char *exclude)
 {
 	struct stat mystat;
-	bool onedev = true;
-
-	if (is_native_overlayfs(path)) {
-		onedev = false;
-	}
 
 	if (lstat(path, &mystat) < 0) {
-		if (errno == ENOENT)
-			return 0;
 		ERROR("%s: failed to stat %s", __func__, path);
 		return -1;
 	}
 
-	return _recursive_rmdir(path, mystat.st_dev, onedev);
+	return _recursive_rmdir_onedev(path, mystat.st_dev, exclude, 0);
+}
+
+static int mount_fs(const char *source, const char *target, const char *type)
+{
+	/* the umount may fail */
+	if (umount(target))
+		WARN("failed to unmount %s : %s", target, strerror(errno));
+
+	if (mount(source, target, type, 0, NULL)) {
+		ERROR("failed to mount %s : %s", target, strerror(errno));
+		return -1;
+	}
+
+	DEBUG("'%s' mounted on '%s'", source, target);
+
+	return 0;
+}
+
+extern void lxc_setup_fs(void)
+{
+	if (mount_fs("proc", "/proc", "proc"))
+		INFO("failed to remount proc");
+
+	/* if we can't mount /dev/shm, continue anyway */
+	if (mount_fs("shmfs", "/dev/shm", "tmpfs"))
+		INFO("failed to mount /dev/shm");
+
+	/* If we were able to mount /dev/shm, then /dev exists */
+	/* Sure, but it's read-only per config :) */
+	if (access("/dev/mqueue", F_OK) && mkdir("/dev/mqueue", 0666)) {
+		DEBUG("failed to create '/dev/mqueue'");
+		return;
+	}
+
+	/* continue even without posix message queue support */
+	if (mount_fs("mqueue", "/dev/mqueue", "mqueue"))
+		INFO("failed to mount /dev/mqueue");
 }
 
 /* borrowed from iproute2 */
@@ -210,6 +250,195 @@ extern int mkdir_p(const char *dir, mode_t mode)
 	return 0;
 }
 
+extern void remove_trailing_slashes(char *p)
+{
+	int l = strlen(p);
+	while (--l >= 0 && (p[l] == '/' || p[l] == '\n'))
+		p[l] = '\0';
+}
+
+static char *copy_global_config_value(char *p)
+{
+	int len = strlen(p);
+	char *retbuf;
+
+	if (len < 1)
+		return NULL;
+	if (p[len-1] == '\n') {
+		p[len-1] = '\0';
+		len--;
+	}
+	retbuf = malloc(len+1);
+	if (!retbuf)
+		return NULL;
+	strcpy(retbuf, p);
+	return retbuf;
+}
+
+#define DEFAULT_VG "lxc"
+#define DEFAULT_THIN_POOL "lxc"
+#define DEFAULT_ZFSROOT "lxc"
+
+const char *lxc_global_config_value(const char *option_name)
+{
+	static const char * const options[][2] = {
+		{ "lxc.bdev.lvm.vg",        DEFAULT_VG      },
+		{ "lxc.bdev.lvm.thin_pool", DEFAULT_THIN_POOL },
+		{ "lxc.bdev.zfs.root",      DEFAULT_ZFSROOT },
+		{ "lxc.lxcpath",            NULL            },
+		{ "lxc.default_config",     NULL            },
+		{ "lxc.cgroup.pattern",     NULL            },
+		{ "lxc.cgroup.use",         NULL            },
+		{ NULL, NULL },
+	};
+
+	/* placed in the thread local storage pool for non-bionic targets */
+#ifdef HAVE_TLS
+	static __thread const char *values[sizeof(options) / sizeof(options[0])] = { 0 };
+#else
+	static const char *values[sizeof(options) / sizeof(options[0])] = { 0 };
+#endif
+
+	/* user_config_path is freed as soon as it is used */
+	char *user_config_path = NULL;
+
+	/*
+	 * The following variables are freed at bottom unconditionally.
+	 * So NULL the value if it is to be returned to the caller
+	 */
+	char *user_default_config_path = NULL;
+	char *user_lxc_path = NULL;
+	char *user_cgroup_pattern = NULL;
+
+	if (geteuid() > 0) {
+		const char *user_home = getenv("HOME");
+		if (!user_home)
+			user_home = "/";
+
+		user_config_path = malloc(sizeof(char) * (22 + strlen(user_home)));
+		user_default_config_path = malloc(sizeof(char) * (26 + strlen(user_home)));
+		user_lxc_path = malloc(sizeof(char) * (19 + strlen(user_home)));
+
+		sprintf(user_config_path, "%s/.config/lxc/lxc.conf", user_home);
+		sprintf(user_default_config_path, "%s/.config/lxc/default.conf", user_home);
+		sprintf(user_lxc_path, "%s/.local/share/lxc/", user_home);
+		user_cgroup_pattern = strdup("%n");
+	}
+	else {
+		user_config_path = strdup(LXC_GLOBAL_CONF);
+		user_default_config_path = strdup(LXC_DEFAULT_CONFIG);
+		user_lxc_path = strdup(LXCPATH);
+		user_cgroup_pattern = strdup(DEFAULT_CGROUP_PATTERN);
+	}
+
+	const char * const (*ptr)[2];
+	size_t i;
+	char buf[1024], *p, *p2;
+	FILE *fin = NULL;
+
+	for (i = 0, ptr = options; (*ptr)[0]; ptr++, i++) {
+		if (!strcmp(option_name, (*ptr)[0]))
+			break;
+	}
+	if (!(*ptr)[0]) {
+		free(user_config_path);
+		free(user_default_config_path);
+		free(user_lxc_path);
+		free(user_cgroup_pattern);
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (values[i]) {
+		free(user_config_path);
+		free(user_default_config_path);
+		free(user_lxc_path);
+		free(user_cgroup_pattern);
+		return values[i];
+	}
+
+	fin = fopen_cloexec(user_config_path, "r");
+	free(user_config_path);
+	if (fin) {
+		while (fgets(buf, 1024, fin)) {
+			if (buf[0] == '#')
+				continue;
+			p = strstr(buf, option_name);
+			if (!p)
+				continue;
+			/* see if there was just white space in front
+			 * of the option name
+			 */
+			for (p2 = buf; p2 < p; p2++) {
+				if (*p2 != ' ' && *p2 != '\t')
+					break;
+			}
+			if (p2 < p)
+				continue;
+			p = strchr(p, '=');
+			if (!p)
+				continue;
+			/* see if there was just white space after
+			 * the option name
+			 */
+			for (p2 += strlen(option_name); p2 < p; p2++) {
+				if (*p2 != ' ' && *p2 != '\t')
+					break;
+			}
+			if (p2 < p)
+				continue;
+			p++;
+			while (*p && (*p == ' ' || *p == '\t')) p++;
+			if (!*p)
+				continue;
+
+			if (strcmp(option_name, "lxc.lxcpath") == 0) {
+				free(user_lxc_path);
+				user_lxc_path = copy_global_config_value(p);
+				remove_trailing_slashes(user_lxc_path);
+				values[i] = user_lxc_path;
+				user_lxc_path = NULL;
+				goto out;
+			}
+
+			values[i] = copy_global_config_value(p);
+			goto out;
+		}
+	}
+	/* could not find value, use default */
+	if (strcmp(option_name, "lxc.lxcpath") == 0) {
+		remove_trailing_slashes(user_lxc_path);
+		values[i] = user_lxc_path;
+		user_lxc_path = NULL;
+	}
+	else if (strcmp(option_name, "lxc.default_config") == 0) {
+		values[i] = user_default_config_path;
+		user_default_config_path = NULL;
+	}
+	else if (strcmp(option_name, "lxc.cgroup.pattern") == 0) {
+		values[i] = user_cgroup_pattern;
+		user_cgroup_pattern = NULL;
+	}
+	else
+		values[i] = (*ptr)[1];
+
+	/* special case: if default value is NULL,
+	 * and there is no config, don't view that
+	 * as an error... */
+	if (!values[i])
+		errno = 0;
+
+out:
+	if (fin)
+		fclose(fin);
+
+	free(user_cgroup_pattern);
+	free(user_default_config_path);
+	free(user_lxc_path);
+
+	return values[i];
+}
+
 char *get_rundir()
 {
 	char *rundir;
@@ -415,6 +644,50 @@ const char** lxc_va_arg_list_to_argv_const(va_list ap, size_t skip)
 	return (const char**)lxc_va_arg_list_to_argv(ap, skip, 0);
 }
 
+FILE *fopen_cloexec(const char *path, const char *mode)
+{
+	int open_mode = 0;
+	int step = 0;
+	int fd;
+	int saved_errno = 0;
+	FILE *ret;
+
+	if (!strncmp(mode, "r+", 2)) {
+		open_mode = O_RDWR;
+		step = 2;
+	} else if (!strncmp(mode, "r", 1)) {
+		open_mode = O_RDONLY;
+		step = 1;
+	} else if (!strncmp(mode, "w+", 2)) {
+		open_mode = O_RDWR | O_TRUNC | O_CREAT;
+		step = 2;
+	} else if (!strncmp(mode, "w", 1)) {
+		open_mode = O_WRONLY | O_TRUNC | O_CREAT;
+		step = 1;
+	} else if (!strncmp(mode, "a+", 2)) {
+		open_mode = O_RDWR | O_CREAT | O_APPEND;
+		step = 2;
+	} else if (!strncmp(mode, "a", 1)) {
+		open_mode = O_WRONLY | O_CREAT | O_APPEND;
+		step = 1;
+	}
+	for (; mode[step]; step++)
+		if (mode[step] == 'x')
+			open_mode |= O_EXCL;
+	open_mode |= O_CLOEXEC;
+
+	fd = open(path, open_mode, 0666);
+	if (fd < 0)
+		return NULL;
+
+	ret = fdopen(fd, mode);
+	saved_errno = errno;
+	if (!ret)
+		close(fd);
+	errno = saved_errno;
+	return ret;
+}
+
 extern struct lxc_popen_FILE *lxc_popen(const char *command)
 {
 	struct lxc_popen_FILE *fp = NULL;
@@ -705,8 +978,8 @@ bool lxc_string_in_list(const char *needle, const char *haystack, char _sep)
 char **lxc_string_split(const char *string, char _sep)
 {
 	char *token, *str, *saveptr = NULL;
-	char sep[2] = {_sep, '\0'};
-	char **tmp = NULL, **result = NULL;
+	char sep[2] = { _sep, '\0' };
+	char **result = NULL;
 	size_t result_capacity = 0;
 	size_t result_count = 0;
 	int r, saved_errno;
@@ -714,7 +987,7 @@ char **lxc_string_split(const char *string, char _sep)
 	if (!string)
 		return calloc(1, sizeof(char *));
 
-	str = alloca(strlen(string) + 1);
+	str = alloca(strlen(string)+1);
 	strcpy(str, string);
 	for (; (token = strtok_r(str, sep, &saveptr)); str = NULL) {
 		r = lxc_grow_array((void ***)&result, &result_capacity, result_count + 1, 16);
@@ -727,14 +1000,7 @@ char **lxc_string_split(const char *string, char _sep)
 	}
 
 	/* if we allocated too much, reduce it */
-	tmp = realloc(result, (result_count + 1) * sizeof(char *));
-	if (!tmp)
-		goto error_out;
-	result = tmp;
-	/* Make sure we don't return uninitialized memory. */
-	if (result_count == 0)
-		*result = NULL;
-	return result;
+	return realloc(result, (result_count + 1) * sizeof(char *));
 error_out:
 	saved_errno = errno;
 	lxc_free_array((void **)result, free);
@@ -1032,6 +1298,31 @@ int detect_shared_rootfs(void)
 	return 0;
 }
 
+bool switch_to_ns(pid_t pid, const char *ns) {
+	int fd, ret;
+	char nspath[MAXPATHLEN];
+
+	/* Switch to new ns */
+	ret = snprintf(nspath, MAXPATHLEN, "/proc/%d/ns/%s", pid, ns);
+	if (ret < 0 || ret >= MAXPATHLEN)
+		return false;
+
+	fd = open(nspath, O_RDONLY);
+	if (fd < 0) {
+		SYSERROR("failed to open %s", nspath);
+		return false;
+	}
+
+	ret = setns(fd, 0);
+	if (ret) {
+		SYSERROR("failed to set process %d to %s of %d.", pid, ns, fd);
+		close(fd);
+		return false;
+	}
+	close(fd);
+	return true;
+}
+
 /*
  * looking at fs/proc_namespace.c, it appears we can
  * actually expect the rootfs entry to very specifically contain
@@ -1071,7 +1362,7 @@ int detect_ramfs_rootfs(void)
 	return 0;
 }
 
-char *on_path(char *cmd) {
+char *on_path(char *cmd, const char *rootfs) {
 	char *path = NULL;
 	char *entry = NULL;
 	char *saveptr = NULL;
@@ -1088,7 +1379,10 @@ char *on_path(char *cmd) {
 
 	entry = strtok_r(path, ":", &saveptr);
 	while (entry) {
-		ret = snprintf(cmdpath, MAXPATHLEN, "%s/%s", entry, cmd);
+		if (rootfs)
+			ret = snprintf(cmdpath, MAXPATHLEN, "%s/%s/%s", rootfs, entry, cmd);
+		else
+			ret = snprintf(cmdpath, MAXPATHLEN, "%s/%s", entry, cmd);
 
 		if (ret < 0 || ret >= MAXPATHLEN)
 			goto next_loop;
@@ -1106,6 +1400,134 @@ next_loop:
 	return NULL;
 }
 
+bool file_exists(const char *f)
+{
+	struct stat statbuf;
+
+	return stat(f, &statbuf) == 0;
+}
+
+/* historically lxc-init has been under /usr/lib/lxc and under
+ * /usr/lib/$ARCH/lxc.  It now lives as $prefix/sbin/init.lxc.
+ */
+char *choose_init(const char *rootfs)
+{
+	char *retv = NULL;
+	const char *empty = "",
+		   *tmp;
+	int ret, env_set = 0;
+	struct stat mystat;
+
+	if (!getenv("PATH")) {
+		if (setenv("PATH", "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin", 0))
+			SYSERROR("Failed to setenv");
+		env_set = 1;
+	}
+
+	retv = on_path("init.lxc", rootfs);
+
+	if (env_set) {
+		if (unsetenv("PATH"))
+			SYSERROR("Failed to unsetenv");
+	}
+
+	if (retv)
+		return retv;
+
+	retv = malloc(PATH_MAX);
+	if (!retv)
+		return NULL;
+
+	if (rootfs)
+		tmp = rootfs;
+	else
+		tmp = empty;
+
+	ret = snprintf(retv, PATH_MAX, "%s/%s/%s", tmp, SBINDIR, "/init.lxc");
+	if (ret < 0 || ret >= PATH_MAX) {
+		ERROR("pathname too long");
+		goto out1;
+	}
+
+	ret = stat(retv, &mystat);
+	if (ret == 0)
+		return retv;
+
+	ret = snprintf(retv, PATH_MAX, "%s/%s/%s", tmp, LXCINITDIR, "/lxc/lxc-init");
+	if (ret < 0 || ret >= PATH_MAX) {
+		ERROR("pathname too long");
+		goto out1;
+	}
+
+	ret = stat(retv, &mystat);
+	if (ret == 0)
+		return retv;
+
+	ret = snprintf(retv, PATH_MAX, "%s/usr/lib/lxc/lxc-init", tmp);
+	if (ret < 0 || ret >= PATH_MAX) {
+		ERROR("pathname too long");
+		goto out1;
+	}
+	ret = stat(retv, &mystat);
+	if (ret == 0)
+		return retv;
+
+	ret = snprintf(retv, PATH_MAX, "%s/sbin/lxc-init", tmp);
+	if (ret < 0 || ret >= PATH_MAX) {
+		ERROR("pathname too long");
+		goto out1;
+	}
+	ret = stat(retv, &mystat);
+	if (ret == 0)
+		return retv;
+
+	/*
+	 * Last resort, look for the statically compiled init.lxc which we
+	 * hopefully bind-mounted in.
+	 * If we are called during container setup, and we get to this point,
+	 * then the init.lxc.static from the host will need to be bind-mounted
+	 * in.  So we return NULL here to indicate that.
+	 */
+	if (rootfs)
+		goto out1;
+
+	ret = snprintf(retv, PATH_MAX, "/init.lxc.static");
+	if (ret < 0 || ret >= PATH_MAX) {
+		WARN("Nonsense - name /lxc.init.static too long");
+		goto out1;
+	}
+	ret = stat(retv, &mystat);
+	if (ret == 0)
+		return retv;
+
+out1:
+	free(retv);
+	return NULL;
+}
+
+int print_to_file(const char *file, const char *content)
+{
+	FILE *f;
+	int ret = 0;
+
+	f = fopen(file, "w");
+	if (!f)
+		return -1;
+	if (fprintf(f, "%s", content) != strlen(content))
+		ret = -1;
+	fclose(f);
+	return ret;
+}
+
+int is_dir(const char *path)
+{
+	struct stat statbuf;
+	int ret = stat(path, &statbuf);
+	if (ret == 0 && S_ISDIR(statbuf.st_mode))
+		return 1;
+	return 0;
+}
+
 /*
  * Given the '-t' template option to lxc-create, figure out what to
  * do.  If the template is a full executable path, use that.  If it
@@ -1140,257 +1562,72 @@ char *get_template_path(const char *t)
 	return tpath;
 }
 
-int null_stdfds(void)
+/*
+ * Sets the process title to the specified title. Note:
+ *   1. this function requires root to succeed
+ *   2. it clears /proc/self/environ
+ *   3. it may not succed (e.g. if title is longer than /proc/self/environ +
+ *      the original title)
+ */
+int setproctitle(char *title)
 {
-	int fd, ret = -1;
+	char buf[2048], *tmp;
+	FILE *f;
+	int i, len, ret = 0;
+	unsigned long arg_start, arg_end, env_start, env_end;
 
-	fd = open("/dev/null", O_RDWR);
-	if (fd < 0)
+	f = fopen_cloexec("/proc/self/stat", "r");
+	if (!f) {
+		return -1;
+	}
+
+	tmp = fgets(buf, sizeof(buf), f);
+	fclose(f);
+	if (!tmp) {
+		return -1;
+	}
+
+	/* Skip the first 47 fields, column 48-51 are ARG_START and
+	 * ARG_END. */
+	tmp = strchr(buf, ' ');
+	for (i = 0; i < 46; i++) {
+		if (!tmp)
+			return -1;
+		tmp = strchr(tmp+1, ' ');
+	}
+
+	if (!tmp)
 		return -1;
 
-	if (dup2(fd, 0) < 0)
-		goto err;
-	if (dup2(fd, 1) < 0)
-		goto err;
-	if (dup2(fd, 2) < 0)
-		goto err;
+	i = sscanf(tmp, "%lu %lu %lu %lu", &arg_start, &arg_end, &env_start, &env_end);
+	if (i != 4) {
+		return -1;
+	}
+
+	/* Include the null byte here, because in the calculations below we
+	 * want to have room for it. */
+	len = strlen(title) + 1;
+
+	/* We're truncating the environment, so we should use at most the
+	 * length of the argument + environment for the title. */
+	if (len > env_end - arg_start) {
+		arg_end = env_end;
+		len = env_end - arg_start;
+	} else {
+		/* Only truncate the environment if we're actually going to
+		 * overwrite part of it. */
+		if (len >= arg_end - arg_start) {
+			env_start = env_end;
+		}
+		arg_end = arg_start + len;
+	}
+
+	strcpy((char*)arg_start, title);
+
+	ret |= prctl(PR_SET_MM, PR_SET_MM_ARG_START,   (long)arg_start, 0, 0);
+	ret |= prctl(PR_SET_MM, PR_SET_MM_ARG_END,     (long)arg_end, 0, 0);
+	ret |= prctl(PR_SET_MM, PR_SET_MM_ENV_START,   (long)env_start, 0, 0);
+	ret |= prctl(PR_SET_MM, PR_SET_MM_ENV_END,     (long)env_end, 0, 0);
 
-	ret = 0;
-err:
-	close(fd);
 	return ret;
 }
-
-/*
- * @path:    a pathname where / replaced with '\0'.
- * @offsetp: pointer to int showing which path segment was last seen.
- *           Updated on return to reflect the next segment.
- * @fulllen: full original path length.
- * Returns a pointer to the next path segment, or NULL if done.
- */
-static char *get_nextpath(char *path, int *offsetp, int fulllen)
-{
-	int offset = *offsetp;
-
-	if (offset >= fulllen)
-		return NULL;
-
-	while (path[offset] != '\0' && offset < fulllen)
-		offset++;
-	while (path[offset] == '\0' && offset < fulllen)
-		offset++;
-
-	*offsetp = offset;
-	return (offset < fulllen) ? &path[offset] : NULL;
-}
-
-/*
- * Check that @subdir is a subdir of @dir.  @len is the length of
- * @dir (to avoid having to recalculate it).
- */
-static bool is_subdir(const char *subdir, const char *dir, size_t len)
-{
-	size_t subdirlen = strlen(subdir);
-
-	if (subdirlen < len)
-		return false;
-	if (strncmp(subdir, dir, len) != 0)
-		return false;
-	if (dir[len-1] == '/')
-		return true;
-	if (subdir[len] == '/' || subdirlen == len)
-		return true;
-	return false;
-}
-
-/*
- * Check if the open fd is a symlink.  Return -ELOOP if it is.  Return
- * -ENOENT if we couldn't fstat.  Return 0 if the fd is ok.
- */
-static int check_symlink(int fd)
-{
-	struct stat sb;
-	int ret = fstat(fd, &sb);
-	if (ret < 0)
-		return -ENOENT;
-	if (S_ISLNK(sb.st_mode))
-		return -ELOOP;
-	return 0;
-}
-
-/*
- * Open a file or directory, provided that it contains no symlinks.
- *
- * CAVEAT: This function must not be used for other purposes than container
- * setup before executing the container's init
- */
-static int open_if_safe(int dirfd, const char *nextpath)
-{
-	int newfd = openat(dirfd, nextpath, O_RDONLY | O_NOFOLLOW);
-	if (newfd >= 0) // was not a symlink, all good
-		return newfd;
-
-	if (errno == ELOOP)
-		return newfd;
-
-	if (errno == EPERM || errno == EACCES) {
-		/* we're not root (cause we got EPERM) so
-		   try opening with O_PATH */
-		newfd = openat(dirfd, nextpath, O_PATH | O_NOFOLLOW);
-		if (newfd >= 0) {
-			/* O_PATH will return an fd for symlinks.  We know
-			 * nextpath wasn't a symlink at last openat, so if fd
-			 * is now a link, then something * fishy is going on
-			 */
-			int ret = check_symlink(newfd);
-			if (ret < 0) {
-				close(newfd);
-				newfd = ret;
-			}
-		}
-	}
-
-	return newfd;
-}
-
-/*
- * Open a path intending for mounting, ensuring that the final path
- * is inside the container's rootfs.
- *
- * CAVEAT: This function must not be used for other purposes than container
- * setup before executing the container's init
- *
- * @target: path to be opened
- * @prefix_skip: a part of @target in which to ignore symbolic links.  This
- * would be the container's rootfs.
- *
- * Return an open fd for the path, or <0 on error.
- */
-static int open_without_symlink(const char *target, const char *prefix_skip)
-{
-	int curlen = 0, dirfd, fulllen, i;
-	char *dup = NULL;
-
-	fulllen = strlen(target);
-
-	/* make sure prefix-skip makes sense */
-	if (prefix_skip && strlen(prefix_skip) > 0) {
-		curlen = strlen(prefix_skip);
-		if (!is_subdir(target, prefix_skip, curlen)) {
-			ERROR("WHOA there - target '%s' didn't start with prefix '%s'",
-				target, prefix_skip);
-			return -EINVAL;
-		}
-		/*
-		 * get_nextpath() expects the curlen argument to be
-		 * on a  (turned into \0) / or before it, so decrement
-		 * curlen to make sure that happens
-		 */
-		if (curlen)
-			curlen--;
-	} else {
-		prefix_skip = "/";
-		curlen = 0;
-	}
-
-	/* Make a copy of target which we can hack up, and tokenize it */
-	if ((dup = strdup(target)) == NULL) {
-		SYSERROR("Out of memory checking for symbolic link");
-		return -ENOMEM;
-	}
-	for (i = 0; i < fulllen; i++) {
-		if (dup[i] == '/')
-			dup[i] = '\0';
-	}
-
-	dirfd = open(prefix_skip, O_RDONLY);
-	if (dirfd < 0)
-		goto out;
-	while (1) {
-		int newfd, saved_errno;
-		char *nextpath;
-
-		if ((nextpath = get_nextpath(dup, &curlen, fulllen)) == NULL)
-			goto out;
-		newfd = open_if_safe(dirfd, nextpath);
-		saved_errno = errno;
-		close(dirfd);
-		dirfd = newfd;
-		if (newfd < 0) {
-			errno = saved_errno;
-			if (errno == ELOOP)
-				SYSERROR("%s in %s was a symbolic link!", nextpath, target);
-			goto out;
-		}
-	}
-
-out:
-	free(dup);
-	return dirfd;
-}
-
-/*
- * Safely mount a path into a container, ensuring that the mount target
- * is under the container's @rootfs.  (If @rootfs is NULL, then the container
- * uses the host's /)
- *
- * CAVEAT: This function must not be used for other purposes than container
- * setup before executing the container's init
- */
-int safe_mount(const char *src, const char *dest, const char *fstype,
-		unsigned long flags, const void *data, const char *rootfs)
-{
-	int srcfd = -1, destfd, ret, saved_errno;
-	char srcbuf[50], destbuf[50]; // only needs enough for /proc/self/fd/<fd>
-	const char *mntsrc = src;
-
-	if (!rootfs)
-		rootfs = "";
-
-	/* todo - allow symlinks for relative paths if 'allowsymlinks' option is passed */
-	if (flags & MS_BIND && src && src[0] != '/') {
-		INFO("this is a relative bind mount");
-		srcfd = open_without_symlink(src, NULL);
-		if (srcfd < 0)
-			return srcfd;
-		ret = snprintf(srcbuf, 50, "/proc/self/fd/%d", srcfd);
-		if (ret < 0 || ret > 50) {
-			close(srcfd);
-			ERROR("Out of memory");
-			return -EINVAL;
-		}
-		mntsrc = srcbuf;
-	}
-
-	destfd = open_without_symlink(dest, rootfs);
-	if (destfd < 0) {
-		if (srcfd != -1) {
-			saved_errno = errno;
-			close(srcfd);
-			errno = saved_errno;
-		}
-		return destfd;
-	}
-
-	ret = snprintf(destbuf, 50, "/proc/self/fd/%d", destfd);
-	if (ret < 0 || ret > 50) {
-		if (srcfd != -1)
-			close(srcfd);
-		close(destfd);
-		ERROR("Out of memory");
-		return -EINVAL;
-	}
-
-	ret = mount(mntsrc, destbuf, fstype, flags, data);
-	saved_errno = errno;
-	if (srcfd != -1)
-		close(srcfd);
-	close(destfd);
-	if (ret < 0) {
-		errno = saved_errno;
-		SYSERROR("Failed to mount %s onto %s", src, dest);
-		return ret;
-	}
-
-	return 0;
-}
